"""Annual date functions visible here are discovered automatically."""

from .easter_funcs import easter, easter_julian, easter_orthodox

__all__ = [
    'easter',
    'easter_julian',
    'easter_orthodox',
]
